/**************************************************************************
 *   tip.c                                                                *
 *                                                                        *
 *   Copyright (C) 1999 Chris Allegretta                                  *
 *   This program is free software; you can redistribute it and/or modify *
 *   it under the terms of the GNU General Public License as published by *
 *   the Free Software Foundation; either version 1, or (at your option)  *
 *   any later version.                                                   *
 *                                                                        *
 *   This program is distributed in the hope that it will be useful,      *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *   GNU General Public License for more details.                         *
 *                                                                        *
 *   You should have received a copy of the GNU General Public License    *
 *   along with this program; if not, write to the Free Software          *
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
 *                                                                        *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>

#include "tip.h"
#include "proto.h"
#include "config.h"

/* What we do when we're all set to exit */
void finish(int sigage)
{
    blank_bottombars();
    wrefresh(bottomwin);
    endwin();
    exit(0);
}

/* Die (gracefully?) */
void die(char *msg, ...)
{
    va_list ap;

    va_start(ap, msg);
    vfprintf(stderr, msg, ap);
    va_end(ap);

    clear();
    refresh();
    resetty();
    endwin();
    printf(msg);
    exit(0);
}

/* Initialize global variables - no better way for now */
void global_init(void)
{
    int i;

    center_x = COLS / 2;
    center_y = LINES / 2;
    current_x = 0;
    current_y = 0;
    editwinrows = LINES - 5;
    editwineob = editwinrows - 1;
    fileage = NULL;
    cutbuffer = NULL;
    current = NULL;
    hblank = malloc(COLS + 1);
    for (i = 0; i <= COLS - 1; i++)
	strcat(hblank, " ");
}

/* Make a copy of a node to a pointer (space will be malloc()ed */
filestruct *copy_node(filestruct * src)
{
    filestruct *dst;

    dst = malloc(sizeof(filestruct));
    dst->data = malloc(strlen(src->data) + 1);

    dst->next = src->next;
    dst->prev = src->prev;

    strcpy(dst->data, src->data);
    dst->lineno = src->lineno;

    return dst;
}

/* Unlink a node from the rest of the struct */
void unlink_node(filestruct * fileptr)
{
    if (fileptr->prev != NULL)
	fileptr->prev->next = fileptr->next;

    if (fileptr->next != NULL)
	fileptr->next->prev = fileptr->prev;
}

void delete_node(filestruct * fileptr)
{
    if (fileptr->data != NULL)
	free(fileptr->data);
    free(fileptr);
}

/* Okay, now let's duplicate a whole struct! */
filestruct *copy_filestruct(filestruct * src)
{
    filestruct *dst, *tmp, *head, *prev;

    head = copy_node(src);
    dst = head;			/* Else we barf on copying just one line :-) */
    tmp = src->next;
    prev = head;

    while (tmp != NULL) {
	dst = copy_node(tmp);
	dst->prev = prev;
	prev->next = dst;

	prev = dst;
	tmp = tmp->next;
    }

    dst->next = NULL;
    return head;
}

/* Free() a single node */
int free_node(filestruct * src)
{
    if (src == NULL)
	return 0;

    if (src->next != NULL)
	free(src->data);
    free(src);
    return 1;
}

int free_filestruct(filestruct * src)
{
    filestruct *fileptr = src;

    if (src == NULL)
	return 0;

    while (fileptr->next != NULL) {
	fileptr = fileptr->next;
	free_node(fileptr->prev);

#ifdef DEBUG
	fprintf(stderr, "free_node(): free'd a node, YAY!\n");
#endif
    }
    free_node(fileptr);
#ifdef DEBUG
    fprintf(stderr, "free_node(): free'd last node.\n");
#endif

    return 1;
}

int renumber_all(void)
{
    filestruct *temp;
    long i = 1;

    for (temp = fileage; temp != NULL; temp = temp->next)
	temp->lineno = i++;

    return 0;
}

int renumber(filestruct * fileptr)
{
    filestruct *temp;

    if (fileptr == NULL || fileptr->prev == NULL || fileptr == fileage) {
	renumber_all();
	return 0;
    }
    for (temp = fileptr; temp != NULL; temp = temp->next)
	temp->lineno = temp->prev->lineno + 1;

    return 0;
}

/* Fix the memory allocation for a string */
void align(char *str)
{
    str = realloc(str, strlen(str) + 1);
}

/* Window I/O */

/* Return the actual place on the screen of current->data[current_x], which 
   should always be > current_x */
int xplustabs(void)
{
    int i, tabs = 0;

    if (current == NULL || current->data == NULL)
	return 0;

    for (i = 1; i <= current_x && current->data[i - 1] != 0; i++) {
	tabs++;

	if (current->data[i - 1] == TIP_CONTROL_I)
	    tabs += 8 - (tabs % 8);
    }

    return tabs;
}

/* Return what current_x should be, given xplustabs() for the line.
   Opposite of xplustabs */
int actual_x(filestruct * fileptr, int xplus)
{
    int i, tot = 0;

    if (fileptr == NULL || fileptr->data == NULL)
	return 0;

    for (i = 1; i + tot <= xplus && fileptr->data[i] != 0; i++)
	if (fileptr->data[i - 1] == TIP_CONTROL_I)
	    tot += 8 - ((i + tot) % 8);

    i--;			/* Im sure there's a good reason why this is needed for
				   it to work, I just cant figure out why =-) */

    return i;
}

/* a strlen with tabs factored in, similar to xplustabs() */
int strlenpt(char *buf)
{
    int i, tabs = 0;

    if (buf == NULL)
	return 0;

    for (i = 1; buf[i] != 0; i++) {
	tabs++;

	if (buf[i - 1] == TIP_CONTROL_I)
	    tabs += 8 - (tabs % 8);
    }

    return tabs;
}


/* resets current_y based on the position of current and puts the cursor at 
   (current_y, current_x) */
void reset_cursor(void)
{
    filestruct *ptr = edittop;

    current_y = 0;

    while (ptr != current && ptr != editbot && ptr->next != NULL) {
	ptr = ptr->next;
	current_y++;
    }

    if (xplustabs() <= COLS - 2)
	wmove(edit, current_y, xplustabs());
    else
	wmove(edit, current_y, xplustabs() % (COLS - 1) + 5);

    wrefresh(edit);
}

void blank_bottombars(void)
{
    int i, j;

    for (j = 1; j <= 2; j++)
	for (i = 0; i <= COLS - 1; i++)
	    mvwaddch(bottomwin, j, i, ' ');

    reset_cursor();
}

void blank_statusbar(void)
{
    int i;

    for (i = 0; i <= COLS - 1; i++)
	mvwaddch(bottomwin, 0, i, ' ');

    reset_cursor();
}

void blank_statusbar_refresh(void)
{
    blank_statusbar();
    wrefresh(bottomwin);
}

void check_statblank(void)
{

    if (statblank > 1)
	statblank--;
    else if (statblank == 1) {
	statblank--;
	blank_statusbar_refresh();
    }
}

/* Get the input from the kb, this should only be called from statusq */

int tipgetstr(char *buf, char *def, shortcut s[], int slen, int start_x)
{
    int kbinput = 0, j = 0, x = 0, xend;
    char inputstr[132], inputbuf[132] = "";

    blank_statusbar();
    mvwaddstr(bottomwin, 0, 0, buf);
    if (strlen(def) > 0)
	waddstr(bottomwin, def);
    wrefresh(bottomwin);

    x = strlen(def) + strlen(buf);

    /* Get the input! */
    if (strlen(def) > 0) {
	strcpy(answer, def);
	strcpy(inputbuf, def);
    }
    /* Go into raw mode so we can actually get ^C, for example */
    raw();

    while ((kbinput = wgetch(bottomwin)) != 13) {
	for (j = 0; j <= slen - 1; j++) {
	    if (kbinput == s[j].val) {
		noraw();
		cbreak();
		strcpy(answer, "");
		return s[j].val;
	    }
	}
	xend = strlen(buf) + strlen(inputbuf);

	if (kbinput >= 32)
	    switch (kbinput) {
	    case KEY_RIGHT:

		if (x < xend)
		    x++;
		wmove(bottomwin, 0, x);
		break;
	    case KEY_BACKSPACE:
	    case KEY_DC:
	    case 127:
		if (strlen(inputbuf) > 0)
		    inputbuf[strlen(inputbuf) - 1] = 0;
		blank_statusbar();
		mvwaddstr(bottomwin, 0, 0, buf);
		waddstr(bottomwin, inputbuf);
	    case KEY_LEFT:
		if (x > strlen(buf))
		    x--;
		wmove(bottomwin, 0, x);
		break;
	    case KEY_UP:
	    case KEY_DOWN:
		break;
	    default:
		strcpy(inputstr, inputbuf);
		inputstr[x - strlen(buf)] = kbinput;
		strcpy(&inputstr[x - strlen(buf) + 1],
		       &inputbuf[x - strlen(buf)]);
		strcpy(inputbuf, inputstr);
		x++;

		mvwaddstr(bottomwin, 0, 0, buf);
		waddstr(bottomwin, inputbuf);
		wmove(bottomwin, 0, x);

#ifdef DEBUG
		fprintf(stderr, "input \'%c\' (%d)\n", kbinput, kbinput);
#endif
	    }
	wrefresh(bottomwin);
    }

    strncpy(answer, inputbuf, 132);

    noraw();
    cbreak();
    if (!strcmp(answer, ""))
	return -2;
    else
	return 0;
}

void horizbar(WINDOW * win, int y)
{
    int i = 0;

    wattron(win, A_REVERSE);
    for (i = 0; i <= COLS - 1; i++)
	mvwaddch(win, y, i, ' ');
    wattroff(win, A_REVERSE);
}

void titlebar(void)
{
    horizbar(topwin, 0);
    wattron(topwin, A_REVERSE);
    mvwaddstr(topwin, 0, 3, VERMSG);
    if (!strcmp(filename, ""))
	mvwaddstr(topwin, 0, center_x - 6, "New Buffer");
    else {
	mvwaddstr(topwin, 0, center_x - 3, "File: ");
	waddstr(topwin, filename);
    }

    if (modified)
	mvwaddstr(topwin, 0, COLS - 10, "Modified");
    wattroff(topwin, A_REVERSE);
    wrefresh(topwin);
    reset_cursor();
}

void onekey(char *keystroke, char *desc)
{
    char description[80];

    snprintf(description, 12, " %-11s", desc);
    wattron(bottomwin, A_REVERSE);
    waddstr(bottomwin, keystroke);
    wattroff(bottomwin, A_REVERSE);
    waddstr(bottomwin, description);
}

void clear_bottomwin(void)
{
    int i;

    for (i = 0; i <= COLS - 1; i++) {
	mvwaddch(bottomwin, 1, i, ' ');
	mvwaddch(bottomwin, 2, i, ' ');
    }
    wrefresh(bottomwin);
}

void bottombars(shortcut s[], int slen)
{
    int i;
    char keystr[10];

    clear_bottomwin();
    wmove(bottomwin, 1, 0);
    for (i = 0; i <= slen - 1; i += 2) {
	sprintf(keystr, "^%c", s[i].val + 64);
	onekey(keystr, s[i].desc);
    }

    wmove(bottomwin, 2, 0);
    for (i = 1; i <= slen - 1; i += 2) {
	sprintf(keystr, "^%c", s[i].val + 64);
	onekey(keystr, s[i].desc);
    }
    wrefresh(bottomwin);

}

/* If modified is not already set, set it and update titlebar */
void set_modified(void)
{
    if (!modified) {
	modified = 1;
	titlebar();
	wrefresh(topwin);
    }
}

/* Just update one line in the edit buffer */
void update_line(filestruct * fileptr)
{
    filestruct *filetmp;
    int line = 0;

    for (filetmp = edittop; filetmp != fileptr && filetmp != editbot;
	 filetmp = filetmp->next)
	line++;

    mvwaddstr(edit, line, 0, filetmp->data);
    reset_cursor();
    wrefresh(edit);

}

void center_cursor(void)
{
    current_y = editwinrows / 2;
    wmove(edit, current_y, current_x);
    wrefresh(edit);
}

/* Refresh the screen without changing the position of lines */
void edit_refresh(void)
{
    int lines = 0, i = 0, col = 0;
    filestruct *temp;

    if (current == NULL)
	return;

    temp = edittop;

    while (lines <= editwinrows - 1 && lines <= totlines &&
	   temp != NULL && temp != filebot) {

	mvwaddstr(edit, lines, 0, hblank);

	if (mark_isset) {

	    if ((temp->lineno > mark_beginbuf->lineno
		 && temp->lineno > current->lineno)
		|| (temp->lineno < mark_beginbuf->lineno
		    && temp->lineno < current->lineno)) {
		mvwaddnstr(edit, lines, 0, temp->data, COLS);
	    } else {
		if (temp != mark_beginbuf && temp != current) {
		    wattron(edit, A_REVERSE);
		    mvwaddstr(edit, lines, 0, temp->data);
		    if (strlenpt(temp->data) > COLS)
			mvwaddch(edit, lines, COLS - 1, '$');

		    wattroff(edit, A_REVERSE);
		}
		/* Special case, we're still on the same line we started marking */
		else if (temp == mark_beginbuf && temp == current) {
		    for (i = 0; i < mark_beginx && i < current_x; i++)
			mvwaddch(edit, lines, i, temp->data[i]);

		    wattron(edit, A_REVERSE);
		    if (i == mark_beginx)
			while (i < current_x) {
			    mvwaddch(edit, lines, i, temp->data[i]);
			    i++;
		    } else
			while (i < mark_beginx) {
			    mvwaddch(edit, lines, i, temp->data[i]);
			    i++;
			}

		    if (strlenpt(temp->data) > COLS)
			mvwaddch(edit, lines, COLS - 1, '$');

		    wattroff(edit, A_REVERSE);
		    mvwaddstr(edit, lines, i, &temp->data[i]);
		} else if (temp == mark_beginbuf) {
		    if (mark_beginbuf->lineno > current->lineno)
			wattron(edit, A_REVERSE);

		    for (i = 0; i <= mark_beginx - 1; i++)
			mvwaddch(edit, lines, i, temp->data[i]);

		    if (strlenpt(temp->data) > COLS)
			mvwaddch(edit, lines, COLS - 1, '$');

		    if (mark_beginbuf->lineno < current->lineno)
			wattron(edit, A_REVERSE);
		    else
			wattroff(edit, A_REVERSE);

		    mvwaddstr(edit, lines, i, &temp->data[i]);
		    if (mark_beginbuf->lineno < current->lineno)
			wattroff(edit, A_REVERSE);
		} else if (temp == current) {
		    if (mark_beginbuf->lineno < current->lineno)
			wattron(edit, A_REVERSE);

		    for (i = 0; i <= current_x - 1; i++)
			mvwaddch(edit, lines, i, temp->data[i]);

		    if (mark_beginbuf->lineno > current->lineno)
			wattron(edit, A_REVERSE);
		    else
			wattroff(edit, A_REVERSE);

		    mvwaddstr(edit, lines, i, &temp->data[i]);
		    if (mark_beginbuf->lineno > current->lineno)
			wattroff(edit, A_REVERSE);
		}
	    }

	} else if (lines == current_y && xplustabs() > COLS - 2) {
	    col = (COLS - 1) * (xplustabs() / (COLS - 1));
	    mvwaddnstr(edit, lines, 1,
		       &temp->data[actual_x(current, col - 4)], COLS);

	    if (strlenpt(&temp->data[col]) > COLS)
		mvwaddch(edit, lines, COLS - 1, '$');
	    mvwaddch(edit, lines, 0, '$');
	} else {
	    mvwaddnstr(edit, lines, 0, temp->data, COLS);

	    if (strlenpt(temp->data) > COLS)
		mvwaddch(edit, lines, COLS - 1, '$');
	}

	temp = temp->next;
	lines++;
    }

    if (temp == filebot) {
	mvwaddstr(edit, lines, 0, hblank);
	mvwaddstr(edit, lines, 0, filebot->data);
	lines++;
    }
    if (lines <= editwinrows - 1)
	while (lines <= editwinrows - 1) {
	    mvwaddstr(edit, lines, i, hblank);
	    lines++;
	}
    editbot = temp;
    wrefresh(edit);
}

/*
 * Nice generic routine to update the edit buffer given a pointer to the
 * file struct =) 
 */
void edit_update(filestruct * fileptr)
{
    int lines = 0, i = 0;
    filestruct *temp;

    if (fileptr == NULL)
	return;

    temp = fileptr;
    while (i <= editwinrows / 2 && temp->prev != NULL) {
	i++;
	temp = temp->prev;
    }
    edittop = temp;

    while (lines <= editwinrows - 1 && lines <= totlines && temp != NULL
	   && temp != filebot) {
	mvwaddstr(edit, lines, 0, hblank);
	mvwaddstr(edit, lines, 0, temp->data);
	temp = temp->next;
	lines++;
    }

    if (temp == filebot) {
	mvwaddstr(edit, lines, 0, hblank);
	mvwaddstr(edit, lines, 0, filebot->data);
	lines++;
	for (i = lines; i <= editwinrows - 1; i++)
	    mvwaddstr(edit, i, 0, hblank);
    }
    editbot = temp;
    wrefresh(edit);
}

/* This function updates current based on where current_y is, reset_cursor 
   does the opposite */
void update_cursor(void)
{
    int i = 0;

    wmove(edit, current_y, current_x);

#ifdef DEBUG
    fprintf(stderr, "Moved to (%d, %d) in edit buffer\n", current_y,
	    current_x);
#endif

    current = edittop;
    while (i <= current_y - 1 && current->next != NULL) {
	current = current->next;
	i++;
    }

#ifdef DEBUG
    fprintf(stderr, "current->data = \"%s\"\n", current->data);
#endif
/*   wrefresh(edit); */

}

void do_first_line(void)
{
    current = fileage;
    placewewant = 0;
    current_x = 0;
    edit_update(current);
}

void do_last_line(void)
{
    current = filebot;
    placewewant = 0;
    current_x = 0;
    edit_update(current);
}

/*
 * Ask a question on the statusbar.  Answer will be stored in answer
 * global.  Returns -1 on aborted enter, -2 on a blank string, and 0
 * otherwise, the valid shortcut key caught, Def is any editable text we
 * want to put up by default.
 */
int statusq(shortcut s[], int slen, char *def, char *msg, ...)
{
    va_list ap;
    char foo[133];
    int ret;

    bottombars(s, slen);

    va_start(ap, msg);
    vsnprintf(foo, 132, msg, ap);
    strncat(foo, ": ", 132);
    va_end(ap);

    wattron(bottomwin, A_REVERSE);
    ret = tipgetstr(foo, def, s, slen, (strlen(foo) + 3));
    wattroff(bottomwin, A_REVERSE);


    switch (ret) {

    case TIP_FIRSTLINE_KEY:
	do_first_line();
	break;
    case TIP_LASTLINE_KEY:
	do_last_line();
	break;
    case TIP_CANCEL_KEY:
	return -1;
    }

    /* Then blank the screen */
    blank_statusbar_refresh();

#ifdef DEBUG
    fprintf(stderr, "I got \"%s\"\n", answer);
#endif

    return ret;
}

/*
 * Ask a simple yes/no question on the statusbar.  Returns 1 for Y, 0 for
 * N, 2 for All (if all is non-zero when passed in) and -1 for abort (^C)
 */
int do_yesno(int all, int leavecursor, char *msg, ...)
{
    va_list ap;
    char foo[133];
    int kbinput, ok = -1;

    /* Write the bottom of the screen */
    clear_bottomwin();
    wattron(bottomwin, A_REVERSE);
    blank_statusbar_refresh();
    wattroff(bottomwin, A_REVERSE);

    wmove(bottomwin, 1, 0);
    onekey(" Y", "Yes");
    if (all)
	onekey(" A", "All");
    wmove(bottomwin, 2, 0);
    onekey(" N", "No");
    onekey("^C", "Cancel");

    va_start(ap, msg);
    vsnprintf(foo, 132, msg, ap);
    va_end(ap);
    wattron(bottomwin, A_REVERSE);
    mvwaddstr(bottomwin, 0, 0, foo);
    wattroff(bottomwin, A_REVERSE);
    wrefresh(bottomwin);

    if (leavecursor == 1)
	reset_cursor();

    raw();

    while (ok == -1) {
	kbinput = wgetch(edit);

	switch (kbinput) {
	case 'Y':
	case 'y':
	    ok = 1;
	    break;
	case 'N':
	case 'n':
	    ok = 0;
	    break;
	case 'A':
	case 'a':
	    if (all)
		ok = 2;
	    break;
	case TIP_CONTROL_C:
	    ok = -2;
	    break;
	}
    }
    noraw();
    cbreak();

    /* Then blank the screen */
    blank_statusbar_refresh();

    if (ok == -2)
	return -1;
    else
	return ok;
}

void statusbar(char *msg, ...)
{
    va_list ap;
    char foo[133];
    int start_x = 0;

    va_start(ap, msg);
    vsnprintf(foo, 132, msg, ap);
    va_end(ap);

    start_x = center_x - strlen(foo) / 2 - 1;

    /* Blank out line */
    blank_statusbar_refresh();

    wmove(bottomwin, 0, start_x);
    wattron(bottomwin, A_REVERSE);

    waddstr(bottomwin, "[ ");
    waddstr(bottomwin, foo);
    waddstr(bottomwin, " ]");
    wattroff(bottomwin, A_REVERSE);
    wrefresh(bottomwin);

    statblank = 25;
}

void total_refresh(void)
{
    int i, j;

    bottombars(main_list, MAIN_LIST_LEN);
    for (i = 0; i != COLS; i++)
	mvwaddch(topwin, 0, i, ' ');
    wrefresh(topwin);
    titlebar();

    for (i = 0; i <= LINES - 1; i++)
	for (j = 0; j <= COLS; j++)
	    mvwaddch(edit, i, j, ' ');
    wrefresh(edit);

    edit_refresh();
/*   reset_cursor(); */
    wrefresh(edit);
}

void previous_line(void)
{
    if (current_y > 0)
	current_y--;
/*   else
   edit_refresh();
 */
/*   reset_cursor(); */
}

/* Dump the current file structure to stderr */
void dump_buffer(filestruct * inptr)
{
#ifdef DEBUG
    filestruct *fileptr;

    if (inptr == fileage)
	fprintf(stderr, "Dumping file buffer to stderr...\n");
    else if (inptr == cutbuffer)
	fprintf(stderr, "Dumping cutbuffer to stderr...\n");
    else
	fprintf(stderr, "Dumping a buffer to stderr...\n");

    fileptr = inptr;
    while (fileptr != NULL) {
	fprintf(stderr, "(%ld) %s\n", fileptr->lineno, fileptr->data);
	fflush(stderr);
	fileptr = fileptr->next;
    }
#endif				/* DEBUG */
}

void dump_buffer_reverse(filestruct * inptr)
{
#ifdef DEBUG
    filestruct *fileptr;

    fileptr = filebot;
    while (fileptr != NULL) {
	fprintf(stderr, "(%ld) %s\n", fileptr->lineno, fileptr->data);
	fflush(stderr);
	fileptr = fileptr->prev;
    }
#endif				/* DEBUG */
}

/* Load file into edit buffer - takes data from file struct */
void load_file(void)
{
    current = fileage;
    wmove(edit, current_y, current_x);
    edit_update(fileage);
    wrefresh(edit);
}

/* What happens when there is no file to open? aiee! */
void new_file(void)
{
    fileage = malloc(sizeof(filestruct));
    fileage->data = malloc(1);
    strcpy(fileage->data, "");
    fileage->prev = NULL;
    fileage->next = NULL;
    fileage->lineno = 1;
    filebot = fileage;
    edittop = fileage;
    editbot = fileage;
    current = fileage;
}

int read_file(char *filename)
{
    long size, totsize = 0, lines = 0, linetemp = 0;
    char input[2];		/* buffer */
    char buf[2048] = "";
    filestruct *fileptr = current, *tmp = NULL;
    int line1ins = 0;

    if (fileptr != NULL && fileptr->prev != NULL) {
	fileptr = fileptr->prev;
	tmp = fileptr;
    } else if (fileptr != NULL && fileptr->prev == NULL) {
	tmp = fileage;
	current = fileage;
	line1ins = 1;
    }

    /* Read the entire file into file struct */
    while ((size = read(file, input, 1)) > 0) {
	linetemp = 0;
	if (input[0] == '\n') {
	    if (line1ins) {

		/* Special case, insert with cursor on 1st line. */
		fileptr = malloc(sizeof(filestruct));
		fileptr->data = malloc(strlen(buf) + 2);
		strcpy(fileptr->data, buf);
		fileptr->prev = NULL;
		fileptr->next = fileage;
		fileptr->lineno = 1;
		line1ins = 0;
		fileage = fileptr;
	    } else if (fileage == NULL) {
		fileage = malloc(sizeof(filestruct));
		fileage->data = malloc(strlen(buf) + 2);
		strcpy(fileage->data, buf);
		fileage->lineno = 1;
		fileage->prev = NULL;
		fileage->next = NULL;
		filebot = fileage;
		fileptr = fileage;
	    } else {
		fileptr->next = malloc(sizeof(filestruct));
		fileptr->next->data = malloc(strlen(buf) + 2);
		strcpy(fileptr->next->data, buf);
		fileptr->next->prev = fileptr;
		fileptr->next->next = NULL;
		fileptr->next->lineno = fileptr->lineno + 1;
		fileptr = fileptr->next;
	    }
	    lines++;
	    strcpy(buf, "\0");
	} else {
	    strncat(buf, input, 1);
	}
	totsize += size;
    }
    if (totsize == 0)
	new_file();

    if (current != NULL) {
	fileptr->next = current;
	current->prev = fileptr;
	renumber(current);
	current_x = 0;
	placewewant = 0;
	edit_update(fileptr);
    } else if (fileptr->next == NULL) {
	filebot = fileptr;
	load_file();
    }
    statusbar("Read %d lines", lines);
    totlines += lines;

    close(file);

    return 1;
}

/* Open the file (and decide if it exists) */
int open_file(char *filename, int insert)
{

    if (!strcmp(filename, "") || stat(filename, &fileinfo) == -1) {
	if (insert) {
	    statusbar("\"%s\" not found", filename);
	    return -1;
	} else {
	    /* We have a new file */
	    statusbar("New File");
	    new_file();
	}
    } else if ((file = open(filename, O_RDONLY)) == -1) {
	statusbar("%s: %s", strerror(errno), filename);
	return -1;
    } else {			/* File is A-OK */

	statusbar("Reading File");
	read_file(filename);
    }

    return 1;
}

int do_insertfile(void)
{
    int i;

    i = statusq(writefile_list, WRITEFILE_LIST_LEN, "",
		"File to insert [from ./] ");
    if (i != -1) {

#ifdef DEBUG
	fprintf(stderr, "filename is %s", answer);
#endif

	i = open_file(answer, 1);

	dump_buffer(fileage);
	set_modified();
	return i;
    } else {
	statusbar("Cancelled");
	return 0;
    }
}

void usage(void)
{
    printf(" Usage: tip -[vwz] +LINE <file>\n");
    printf(" -v: Print version information and exit\n");
    printf(" -w: Don't wrap long lines\n");
    printf(" -z: Enable suspend\n");
    exit(0);
}

void version(void)
{
    printf(" tip version %s by Chris Allegretta\n", VERSION);
}

void page_down(void)
{
    if (editbot->next != NULL && editbot->next != filebot) {
	edit_update(editbot->next);
	center_cursor();
    } else if (editbot != filebot) {
	edit_update(editbot);
	center_cursor();
    } else
	while (current != filebot) {
	    current = current->next;
	    current_y++;
	}

    update_cursor();
}

filestruct *make_new_node(filestruct * prevnode)
{
    filestruct *newnode;

    newnode = malloc(sizeof(filestruct));
    newnode->data = NULL;

    newnode->prev = prevnode;
    newnode->next = NULL;

    if (prevnode != NULL)
	newnode->lineno = prevnode->lineno + 1;

    return newnode;
}

void add_to_cutbuffer(filestruct * inptr)
{
    filestruct *tmp;

#ifdef DEBUG
    fprintf(stderr, "add_to_cutbuffer called with inptr->data = %s\n",
	    inptr->data);
#endif

    tmp = cutbuffer;
    if (cutbuffer == NULL) {
	cutbuffer = inptr;
	inptr->next = NULL;
	inptr->prev = NULL;
	return;
    } else {
	while (tmp->next != NULL)
	    tmp = tmp->next;
    }

    tmp->next = inptr;
    inptr->prev = tmp;
    inptr->next = NULL;
    cutbottom = inptr;
    totlines--;
}

void cut_marked_segment(filestruct * top, int top_x, filestruct * bot,
			int bot_x)
{
    filestruct *tmp, *next;
    char *tmpstr;

    tmp = copy_node(top);
    tmpstr = malloc(strlen(&top->data[top_x]) + 1);
    strcpy(tmpstr, &top->data[top_x]);
    free(tmp->data);
    tmp->data = tmpstr;

    do {
	next = tmp->next;
	add_to_cutbuffer(tmp);
	tmp = next;
    }
    while (next != bot && next != NULL);

    dump_buffer(cutbuffer);
    if (next == NULL)
	return;
    /* Now, paste bot[bot_x + 1] into top[top_x] */
    tmpstr = malloc(strlen(top->data) + strlen(&bot->data[bot_x + 1]));
    strncpy(tmpstr, top->data, top_x);
    strcpy(&tmpstr[top_x], &bot->data[bot_x + 1]);
    free(top->data);
    top->data = tmpstr;

    bot->data[bot_x] = 0;
    align(bot->data);
    next = bot->next;
    add_to_cutbuffer(bot);
    top->next = next;
    if (next != NULL)
	next->prev = top;

    dump_buffer(cutbuffer);
    renumber(top);
    current = top;
    current_x = top_x;
    if (top->lineno < edittop->lineno)
	edit_update(top);
}


void do_cut_text(filestruct * fileptr)
{
    filestruct *tmp;
    char *tmpstr;
    int newsize;

    if (fileptr == NULL || fileptr->data == NULL ||
	(filebot == fileptr && !strcmp(fileptr->data, "")))
	return;

    tmp = fileptr->next;

    if (!keep_cutbuffer || mark_isset) {
	free_filestruct(cutbuffer);
	cutbuffer = NULL;
#ifdef DEBUG
	fprintf(stderr, "Blew away cutbuffer =)\n");
#endif
    }
    if (mark_isset) {
	if (current->lineno == mark_beginbuf->lineno) {
	    tmp = copy_node(current);
	    newsize = abs(strlen(&current->data[mark_beginx]) -
			  strlen(&current->data[current_x]));

	    tmpstr = malloc(newsize);
	    if (current_x < mark_beginx) {
		strncpy(tmpstr, &current->data[current_x], newsize);
		memmove(&current->data[current_x],
			&current->data[mark_beginx],
			strlen(&current->data[mark_beginx] - newsize));
	    } else {
		strncpy(tmpstr, &current->data[mark_beginx], newsize);
		memmove(&current->data[mark_beginx],
			&current->data[current_x],
			strlen(&current->data[current_x] - newsize));
		current_x = mark_beginx;
		update_cursor();
	    }
	    tmpstr[newsize] = 0;
	    tmp->data = tmpstr;
	    add_to_cutbuffer(tmp);
	    dump_buffer(cutbuffer);
	    align(current->data);
	} else if (current->lineno < mark_beginbuf->lineno)
	    cut_marked_segment(current, current_x, mark_beginbuf,
			       mark_beginx);
	else
	    cut_marked_segment(mark_beginbuf, mark_beginx, current,
			       current_x);

	placewewant = xplustabs();
	mark_isset = 0;
	marked_cut = 1;
	set_modified();
	return;
    } else if (fileptr == fileage) {
	if (fileptr->next != NULL) {
	    fileptr = fileptr->next;
	    tmp = fileptr;
	    fileage = fileptr;
	    add_to_cutbuffer(fileptr->prev);
	    fileptr->prev = NULL;
	    edit_update(fileage);
	} else {
	    add_to_cutbuffer(fileptr);
	    fileage = make_new_node(NULL);
	    fileage->data = malloc(1);
	    strcpy(fileage->data, "");
	}
    } else {
	if (fileptr->prev != NULL)
	    fileptr->prev->next = fileptr->next;

	if (fileptr->next != NULL)
	    (fileptr->next)->prev = fileptr->prev;
	else
	    filebot = fileptr->prev;

	add_to_cutbuffer(fileptr);
    }

    if (tmp != NULL)
	current = tmp;
    else {			/* FIXME - wrong */

	tmp = make_new_node(filebot);
	tmp->data = malloc(1);
	strcpy(tmp->data, "");
	filebot->next = tmp;
	current = tmp;
	filebot = tmp;
    }

    if (fileptr == edittop)
	edittop = current;

    edit_refresh();

    dump_buffer(cutbuffer);
    reset_cursor();

    set_modified();
    keep_cutbuffer = 1;
    marked_cut = 0;
    current_x = 0;
    placewewant = 0;
    update_cursor();
    renumber(tmp);
}

void do_uncut_text(filestruct * fileptr)
{
    filestruct *tmp = fileptr, *newbuf, *newend;
    char *tmpstr, *tmpstr2;
    int i;

#ifdef DEBUG
    fprintf(stderr, "do_uncut_text called with fileptr->data = %s\n",
	    fileptr->data);
#endif

    if (cutbuffer == NULL || fileptr == NULL)
	return;			/* AIEEEEEEEEEEEE */

    newbuf = copy_filestruct(cutbuffer);
    /* Make newend = last element in newbuf */
    for (newend = newbuf; newend->next != NULL && newend != NULL;
	 newend = newend->next) totlines++;

    /* Hook newbuf into fileptr */
    if (marked_cut == 1) {
	/* If there's only one line in the cutbuffer */
	if (cutbuffer->next == NULL) {
	    tmpstr =
		malloc(strlen(current->data) + strlen(cutbuffer->data) +
		       1);
	    strncpy(tmpstr, current->data, current_x);
	    strcpy(&tmpstr[current_x], cutbuffer->data);
	    strcat(tmpstr, &current->data[current_x]);
	    free(current->data);
	    current->data = tmpstr;
	    current_x += strlen(cutbuffer->data);

	    placewewant = xplustabs();
	    update_cursor();
	    renumber(current);
	} else {		/* yuck */

	    tmp = current->next;
	    /* New beginning */
	    tmpstr = malloc(current_x + strlen(newbuf->data) + 1);
	    strncpy(tmpstr, current->data, current_x);
	    strcpy(&tmpstr[current_x], newbuf->data);

	    /* New end */
	    tmpstr2 = malloc(strlen(newend->data) +
			     strlen(&current->data[current_x]) + 1);
	    strcpy(tmpstr2, newend->data);
	    strcat(tmpstr2, &current->data[current_x]);

	    free(current->data);
	    current->data = tmpstr;
	    current->next = newbuf->next;
	    newbuf->next->prev = current;
	    delete_node(newbuf);

	    current_x = strlen(newend->data);
	    placewewant = xplustabs();
	    free(newend->data);
	    newend->data = tmpstr2;

	    newend->next = tmp;
	    if (tmp != NULL)
		tmp->prev = newend;

	    i = editbot->lineno;
	    renumber(current);
	    current = newend;
	    if (i <= newend->lineno)
		edit_update(current);
/*         update_cursor(); */
	}

	dump_buffer(fileage);
	dump_buffer(cutbuffer);
	set_modified();
	return;

    } else if (fileptr != fileage) {
	tmp = fileptr->prev;
	tmp->next = newbuf;
	newbuf->prev = tmp;
    } else
	fileage = newbuf;

    /* This is so uncutting at the top of the buffer will work => */
    if (current_y == 0)
	edittop = newbuf;

    /* Connect the end of the buffer to the filestruct */
    newend->next = fileptr;
    fileptr->prev = newend;

    i = editbot->lineno;
    renumber(newbuf);
    if (i < newend->lineno)
	edit_update(fileptr);

/*    dump_buffer(cutbuffer);
   dump_buffer(fileage); */
    dump_buffer_reverse(fileptr);

    set_modified();
}

void do_mark()
{
    if (!mark_isset) {
	statusbar("Mark Set");
	mark_isset = 1;
	mark_beginbuf = current;
	mark_beginx = current_x;
    } else {
	statusbar("Mark UNset");
	mark_isset = 0;
	mark_beginbuf = NULL;
	mark_beginx = 0;
    }
}

void do_early_abort(void)
{
    blank_statusbar_refresh();
    bottombars(main_list, MAIN_LIST_LEN);
/*    reset_cursor(); */
}

/* Set up the system variables for a search or replace.  Returns -1 on
   abort, 0 on success, and 1 on rerun calling program */
int search_init(void)
{
    int i;

    if (strcmp(last_search, "")) {	/* There's a previous search stored */
	if (case_sensitive)
	    i = statusq(whereis_list, WHEREIS_LIST_LEN, "",
			"Case Sensitive Search [%s]", last_search);
	else
	    i = statusq(whereis_list, WHEREIS_LIST_LEN, "", "Search [%s]",
			last_search);

	if (i == -1) {		/* Aborted enter */
	    strncpy(answer, last_search, 132);
	    statusbar("Search Cancelled");
	    return -1;
	} else if (i == -2) {	/* Same string */
	    strncpy(answer, last_search, 132);
	} else if (i == 0) {	/* They actually entered something */
	    strncpy(last_search, answer, 132);

	    /* Blow away last_replace because they entered a new search
	       string....uh, right? =) */
	    strcpy(last_replace, "");
	} else if (i == TIP_CASE_KEY) {	/* They asked for case sensitivity */
	    case_sensitive = 1 - case_sensitive;
	    return 1;
	} else {		/* First page, last page, for example could get here */

	    do_early_abort();
	    return -3;
	}
    } else {			/* last_search is empty */

	if (case_sensitive)
	    i = statusq(whereis_list, WHEREIS_LIST_LEN, "",
			"Case Sensititve Search");
	else
	    i = statusq(whereis_list, WHEREIS_LIST_LEN, "", "Search");
	if (i < 0) {
	    statusbar("Search Cancelled");
	    reset_cursor();
	    return -1;
	} else if (i == 0)	/* They entered something new */
	    strncpy(last_search, answer, 132);
	else if (i == TIP_CASE_KEY) {	/* They want it case sensitive */
	    case_sensitive = 1 - case_sensitive;
	    return 1;
	} else {		/* First line key, etc. */

	    do_early_abort();
	    return -3;
	}
    }

    return 0;
}

filestruct *findnextstr(int quiet, filestruct * begin, char *needle)
{
    filestruct *fileptr;
    char *searchstr, *found = NULL, *tmp;

    fileptr = current;

    searchstr = &current->data[current_x + 1];
    /* Look for searchstr until EOF */
    while (fileptr != NULL &&
	   (found = strstrwrapper(searchstr, needle)) == NULL) {
	fileptr = fileptr->next;

	if (fileptr == begin)
	    return NULL;

	if (fileptr != NULL)
	    searchstr = fileptr->data;
    }

    /* If we're not at EOF, we found an instance */
    if (fileptr != NULL) {
	current = fileptr;
	current_x = 0;
	for (tmp = fileptr->data; tmp != found; tmp++)
	    current_x++;

	edit_update(current);
	reset_cursor();
    } else {			/* We're at EOF, go back to the top, once */

	fileptr = fileage;

	while (fileptr != current && fileptr != begin &&
	       (found = strstrwrapper(fileptr->data, needle)) == NULL)
	    fileptr = fileptr->next;

	if (fileptr == begin)
	    return NULL;

	if (fileptr != current) {	/* We found something */
	    current = fileptr;
	    current_x = 0;
	    for (tmp = fileptr->data; tmp != found; tmp++)
		current_x++;

	    edit_update(current);
	    reset_cursor();

	    if (!quiet)
		statusbar("Search Wrapped");
	} else {		/* Nada */

	    if (!quiet)
		statusbar("Search string not found");
	    return NULL;
	}
    }

    return fileptr;
}

/* Search for a string */
void do_search(void)
{
    int i;
    filestruct *fileptr = current;

    if ((i = search_init()) == -1) {
	current = fileptr;
	return;
    } else if (i == -3) {
	return;
    } else if (i == 1) {
	do_search();
	return;
    }

    findnextstr(0, current, answer);
}

void print_replaced(int num)
{
    if (num > 1)
	statusbar("Replaced %d occurences", num);
    else if (num == 1)
	statusbar("Replaced 1 occurence");
}

/* Search for a string */
void do_replace(void)
{
    int i, j, replaceall = 0, numreplaced = 0, beginx;
    filestruct *fileptr, *begin;
    char *tmp, *copy, prevanswer[132] = "";

    if ((i = search_init()) == -1) {
	statusbar("Replace Cancelled");
	return;
    } else if (i == 1) {
	do_replace();
	return;
    } else if (i == -3) {
	return;
    }
    strncpy(prevanswer, answer, 132);

    if (strcmp(last_replace, "")) {	/* There's a previous replace str */
	i = statusq(replace_list, REPLACE_LIST_LEN, "",
		    "Replace with [%s]", last_replace);

	if (i == -1) {		/* Aborted enter */
	    strncpy(answer, last_replace, 132);
	    statusbar("Replace Cancelled");
	    return;
	} else if (i == 0)	/* They actually entered something */
	    strncpy(last_replace, answer, 132);
	else if (i == TIP_CASE_KEY) {	/* They asked for case sensitivity */
	    case_sensitive = 1 - case_sensitive;
	    do_replace();
	    return;
	} else {		/* First page, last page, for example could get here */

	    do_early_abort();
	    return;
	}
    } else {			/* last_search is empty */

	i = statusq(replace_list, REPLACE_LIST_LEN, "", "Replace with");
	if (i == -1) {
	    statusbar("Replace Cancelled");
	    reset_cursor();
	    return;
	} else if (i == 0)	/* They entered something new */
	    strncpy(last_replace, answer, 132);
	else if (i == TIP_CASE_KEY) {	/* They want it case sensitive */
	    case_sensitive = 1 - case_sensitive;
	    do_replace();
	    return;
	} else {		/* First line key, etc. */

	    do_early_abort();
	    return;
	}
    }

    begin = current;
    beginx = current_x;
    while (1) {

	if (replaceall)
	    fileptr = findnextstr(1, begin, prevanswer);
	else
	    fileptr = findnextstr(0, begin, prevanswer);

	if (fileptr == NULL) {
	    current = begin;
	    current_x = beginx;
	    edit_update(current);
	    print_replaced(numreplaced);
	    return;
	}
	/* If we're here, we've found the search string */
	if (!replaceall)
	    i = do_yesno(1, 1, "Replace this instance?");

	if (i == 1 || replaceall) {	/* Yes, replace it!!!! */

	    /* FIXME - lots of ugly code */
	    copy = malloc(strlen(current->data) - strlen(last_search) +
			  strlen(last_replace) + 1);

	    strncpy(copy, current->data, current_x);
	    copy[current_x] = 0;

	    strcat(copy, last_replace);

	    for (j = 1, tmp = current->data; j <=
		 (strlen(last_search) + current_x) && *tmp != 0; j++)
		tmp++;

	    if (*tmp != 0)
		strcat(copy, tmp);

	    tmp = current->data;
	    current->data = copy;

	    free(tmp);

	    edit_refresh();
	    set_modified();
	    numreplaced++;
	} else if (i == 2)	/* replace all, aieeeeeeeeeeeeee */
	    replaceall = 1;
	else if (i == -1)	/* Abort, else do nothing and continue loop */
	    break;
    }

    print_replaced(numreplaced);
}


/* What happens when we want to go past the bottom of the buffer */
int do_down(void)
{
    if (current->next != NULL) {
	if (placewewant > 0)
	    current_x = actual_x(current->next, placewewant);

	if (current_x > strlen(current->next->data) - 1)
	    current_x = strlen(current->next->data) - 1;
    } else
	return 0;

    if (current_y < editwineob && current != editbot)
	current_y++;
    else
	page_down();

    return 1;
}

int page_up(void)
{
    if (edittop != fileage) {
	edit_update(edittop);
	center_cursor();
    } else
	current_y = 0;

    update_cursor();

    return 1;
}

void do_up(void)
{
    if (current->prev != NULL) {
	if (placewewant > 0)
	    current_x = actual_x(current->prev, placewewant);

	if (current_x > strlen(current->prev->data))
	    current_x = strlen(current->prev->data);
    }
    if (current_y > 0)
	current_y--;
    else
	page_up();

    wrefresh(edit);
}

void do_right(void)
{
    if (current_x < strlen(current->data)) {
	current_x++;
    } else {
	if (do_down())
	    current_x = 0;
    }

    placewewant = xplustabs();
}

void do_left(void)
{
    if (current_x > 0)
	current_x--;
    else if (current != fileage) {
	current_x = strlen(current->prev->data);
	do_up();
    } else
	statusbar("Beep!");

    placewewant = xplustabs();
}

void delete_buffer(filestruct * inptr)
{
    if (inptr != NULL) {
	delete_buffer(inptr->next);
	free(inptr->data);
	free(inptr);
    }
}

void do_backspace(void)
{
    filestruct *previous;

    if (current_x != 0) {
	/* Let's get dangerous */
	memmove(&current->data[current_x - 1], &current->data[current_x],
		strlen(current->data) - current_x + 2);
#ifdef DEBUG
	fprintf(stderr, "current->data now = \"%s\"\n", current->data);
#endif
	align(current->data);
	do_left();
    } else {
	if (current == fileage)
	    return;		/* Can't delete past top of file */

	previous = current->prev;
	current_x = strlen(previous->data);
	previous->data = realloc(previous->data,
				 strlen(previous->data) +
				 strlen(current->data) + 1);
	strcat(previous->data, current->data);
	if (current->next != NULL)
	    current->next->prev = current->prev;
	previous->next = current->next;

	unlink_node(current);
	delete_node(current);
	if (current == edittop)
	    page_up();
	current = previous;
	previous_line();
	totlines--;

	renumber(current);

#ifdef DEBUG
	fprintf(stderr, "After, data = \"%s\"\n", current->data);
#endif

    }

    set_modified();
    keep_cutbuffer = 0;
}

void do_delete(void)
{
    if (current_x != strlen(current->data)) {
	/* Let's get dangerous */
	memmove(&current->data[current_x], &current->data[current_x + 1],
		strlen(current->data) - current_x + 1);
#ifdef DEBUG
	fprintf(stderr, "current->data now = \"%s\"\n", current->data);
#endif
	align(current->data);

    } else if (current->next != NULL) {
	current_x = 0;
	current = current->next;
	do_backspace();
    } else
	return;

    set_modified();
    keep_cutbuffer = 0;
}

/* Someone hits return *gasp!* */
void do_enter(filestruct * inptr)
{
    filestruct *new;
    char *tmp;

    new = make_new_node(inptr);

    tmp = &current->data[current_x];
    new->data = malloc(strlen(tmp) + 1);
    strcpy(new->data, tmp);
    *tmp = 0;

    new->next = inptr->next;
    new->prev = inptr;
    inptr->next = new;
    if (new->next != NULL)
	new->next->prev = new;
    else
	filebot = new;

    renumber(current);
    current = new;
    current_x = 0;
    align(current->data);

    if (current_y == editwinrows - 1)
	edit_update(current);

    totlines++;
    set_modified();
}

/* Actually wrap a line, called by check_wrap() */
void do_wrap(filestruct * inptr)
{
    int i, j;
    char *tmp, *foo;

    i = actual_x(inptr, COLS - 1);
    if (inptr->data[i] == ' ')
	while (inptr->data[i] == ' ' && i != 0)
	    i--;

    while (inptr->data[i] != ' ' && i != 0)
	i--;
    i++;

    if (samelinewrap) {
	tmp = &current->data[i];
	foo = malloc(strlen(tmp) + strlen(current->next->data) + 1);
	strcpy(foo, tmp);
	strcpy(&foo[strlen(tmp)], current->next->data);
	free(current->next->data);
	current->next->data = foo;
	*tmp = 0;
	if (current_x >= i) {
	    current_x = current_x - i;
	    current = current->next;
	}
	align(current->next->data);
    } else {
	j = current_x;
	current_x = i;
	do_enter(current);
	if (j > i) {
	    current_x = j - i;
	    samelinewrap = 0;
	} else {
	    current_x = j;
	    current = current->prev;
	    samelinewrap = 1;
	}
    }

}

/* Check to see if we've just caused the line to wrap to a new line */
void check_wrap(filestruct * inptr)
{

#ifdef DEBUG
    fprintf(stderr, "check_wrap called with inptr->data=\"%s\"\n",
	    inptr->data);
#endif

    if ((int) strlenpt(inptr->data) <= COLS)
	return;
    else
	do_wrap(inptr);
}

void do_gotoline(void)
{
    long line, i = 1, j = 0;
    filestruct *fileptr;

    j = statusq(replace_list, REPLACE_LIST_LEN, "", "Enter line number");
    if (j == -1) {
	statusbar("Aborted");
	reset_cursor();
	return;
    } else if (j != 0) {
	do_early_abort();
	return;
    }
    if (!strcmp(answer, "$")) {
	current = filebot;
	current_x = 0;
	edit_update(current);
	reset_cursor();
	return;
    }
    line = atoi(answer);
    /* Bounds check */
    if (line <= 0) {
	statusbar("Come on, be reasonable");
	return;
    }
    if (line > totlines) {
	statusbar("Only %d lines available, skipping to last line",
		  filebot->lineno);
	current = filebot;
	current_x = 0;
	edit_update(current);
	reset_cursor();
    } else {
	for (fileptr = fileage; fileptr != NULL && i < line; i++)
	    fileptr = fileptr->next;

	current = fileptr;
	current_x = 0;
	edit_update(current);
	reset_cursor();
    }

}

void wrap_reset(void)
{
    samelinewrap = 0;
}

int write_file(char *name)
{
    long size, lineswritten = 0;
    filestruct *fileptr;

    titlebar();
    fileptr = fileage;

    if ((file = open(name, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR
		     | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)) == -1) {
	statusbar("Could not open file for writing: %s", strerror(errno));
	return -1;
    }
    dump_buffer(fileage);
    while (fileptr != NULL) {
	size = write(file, fileptr->data, strlen(fileptr->data));
	if (size == -1) {
	    statusbar("Could not open file for writing: %s",
		      strerror(errno));
	    return -1;
	} else {
#ifdef DEBUG
	    fprintf(stderr, "Wrote >%s\n", fileptr->data);
#endif
	}
	write(file, "\n", 1);

	fileptr = fileptr->next;
	lineswritten++;
    }

    strncpy(filename, name, 132);
    statusbar("Wrote %d lines", lineswritten);
    modified = 0;
    titlebar();
    return 1;
}

int do_writeout(void)
{
    int i;

    i = statusq(writefile_list, WRITEFILE_LIST_LEN, filename,
		"File Name to write");
    if (i != -1) {

#ifdef DEBUG
	fprintf(stderr, "filename is %s", answer);
#endif

	i = write_file(answer);

	return i;
    } else {
	statusbar("Cancelled");
	return 0;
    }
}


void do_exit(void)
{
    int i;

    if (!modified)
	finish(0);

    i =
	do_yesno(0, 0,
		 "Save modified buffer (ANSWERING \"No\" WILL DESTROY CHANGES) ? ");

#ifdef DEBUG
    dump_buffer(fileage);
#endif

    if (i == 1) {
	if (do_writeout() > 0)
	    finish(0);

    } else if (i == 0)
	finish(0);
    else
	statusbar("Cancelled");

    bottombars(main_list, MAIN_LIST_LEN);
}

int main(int argc, char *argv[])
{
    char optchr;
    int kbinput;		/* Input from keyboard */

    while ((optchr = getopt(argc, argv, "h?wvz")) != EOF) {
	switch (optchr) {
	case 'h':
	case '?':
	    usage();
	case 'v':
	    version();
	    exit(0);
	case 'z':
	    suspend = 1;
	    break;
	case 'w':
	    no_wrap = 1;
	    break;
	default:
	    usage();
	}
    }

    if (argc == 1)
	strcpy(filename, "");
    else
	strncpy(filename, argv[argc - 1], 132);

    initscr();
    savetty();
    nonl();
    cbreak();
    noecho();
    timeout(0);

    /* Set up some global variables */
    global_init();

    signal(SIGINT, SIG_IGN);
    if (!suspend)
	signal(SIGTSTP, SIG_IGN);

#ifdef DEBUG
    fprintf(stderr, "Main: set up windows\n");
#endif

    /* Setup up the main text window */
    edit = newwin(editwinrows, COLS, 2, 0);
    keypad(edit, TRUE);

    /* And the other windows */
    topwin = newwin(2, COLS, 0, 0);
    bottomwin = newwin(3, COLS, LINES - 3, 0);
    keypad(bottomwin, TRUE);

#ifdef DEBUG
    fprintf(stderr, "Main: bottom win\n");
#endif
    /* Set up up bottom of window */
    bottombars(main_list, MAIN_LIST_LEN);

#ifdef DEBUG
    fprintf(stderr, "Main: open file\n");
#endif

    titlebar();
    if (argc == 1)
	new_file();
    else
	open_file(filename, 0);

    edit_refresh();
    wmove(edit, 0, 0);
    wrefresh(edit);

    while (1) {
	kbinput = wgetch(edit);
	switch (kbinput) {
	case TIP_EXIT_KEY:
	    do_exit();
	    break;
	case TIP_WRITEOUT_KEY:
	    do_writeout();
	    bottombars(main_list, MAIN_LIST_LEN);
	    break;
	case TIP_INSERTFILE_KEY:
	    wrap_reset();
	    do_insertfile();
	    keep_cutbuffer = 0;
	    bottombars(main_list, MAIN_LIST_LEN);
	    break;
	case TIP_GOTO_KEY:
	    wrap_reset();
	    do_gotoline();
	    keep_cutbuffer = 0;
	    bottombars(main_list, MAIN_LIST_LEN);
	    break;
	case TIP_WHEREIS_KEY:
	    wrap_reset();
	    do_search();
	    keep_cutbuffer = 0;
	    bottombars(main_list, MAIN_LIST_LEN);
	    wrefresh(bottomwin);
	    break;
	case TIP_CUT_KEY:
	    do_cut_text(current);
	    break;
	case TIP_CONTROL_A:
	    current_x = 0;
	    placewewant = 0;
	    break;
	case TIP_CONTROL_E:
	    current_x = strlen(current->data);
	    placewewant = xplustabs();
	    break;
	case TIP_PREVPAGE_KEY:
	    wrap_reset();
	    current_x = 0;
	    page_up();
	    keep_cutbuffer = 0;
	    check_statblank();
	    break;
	case TIP_NEXTPAGE_KEY:
	    wrap_reset();
	    current_x = 0;
	    page_down();
	    keep_cutbuffer = 0;
	    check_statblank();
	    break;
	case TIP_UNCUT_KEY:
	    wrap_reset();
	    do_uncut_text(current);
	    keep_cutbuffer = 0;
	    break;
	case 30:		/* ^^ */
	    do_mark();
	    break;
	case TIP_SPELL_KEY:
#ifdef DEBUG
	    fprintf(stderr, "To Spell - not yet implemented\n");
#endif
	    keep_cutbuffer = 0;
	    break;
	case TIP_REPLACE_KEY:
	    do_replace();
	    keep_cutbuffer = 0;
	    bottombars(main_list, MAIN_LIST_LEN);
	    break;
	case KEY_UP:
	    wrap_reset();
	    do_up();
	    update_cursor();
	    keep_cutbuffer = 0;
	    check_statblank();
	    break;
	case KEY_DOWN:
	    wrap_reset();
	    do_down();
	    update_cursor();
	    keep_cutbuffer = 0;
	    check_statblank();
	    break;
	case KEY_LEFT:
	    do_left();
	    update_cursor();
	    keep_cutbuffer = 0;
	    check_statblank();
	    break;
	case KEY_RIGHT:
	    do_right();
	    update_cursor();
	    keep_cutbuffer = 0;
	    check_statblank();
	    break;
	case KEY_BACKSPACE:
	case 127:
	    do_backspace();
	    break;
	case KEY_DC:
	    do_delete();
	    break;
	case TIP_REFRESH_KEY:
	    total_refresh();
	    break;
	case TIP_CONTROL_M:	/* Enter (^M) - FIXME - must be more 
				   things bound to enter key */
	    do_enter(current);
	    break;
	default:
	    if (kbinput >= 1 && kbinput <= 26 && kbinput != TIP_CONTROL_I)
		break;		/* Unhandled control code */

	    /* More dangerousness fun =) */
	    current->data =
		realloc(current->data, strlen(current->data) + 2);
	    memmove(&current->data[current_x + 1],
		    &current->data[current_x],
		    strlen(current->data) - current_x + 1);
	    current->data[current_x] = kbinput;
	    update_line(current);
	    do_right();
	    update_cursor();
	    if (!no_wrap)
		check_wrap(current);
	    set_modified();
	    check_statblank();
	    keep_cutbuffer = 0;
	}
	edit_refresh();
	reset_cursor();

    }

    getchar();
    finish(0);

}
